#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  RS_FKM_CL.py
#
#  Random Swap and Fast Kmeans clustering implemented in OpenCL
#  Note: The implementation is quite naive and could be optimized per 
#        hardware
import numpy as np
import pyopencl as cl
import random as r
import argparse
from time import time

# Name of the file containing OpenCL kernel source code
KERNEL_FILE = "rs_fkm.cl"

def compute_TSE(centroids, datapoints):
    """ Calculate total squared error """
    distances = np.expand_dims(datapoints, 1)-centroids
    distances = np.sum(distances**2, axis=2)
    partition = np.argmin(distances,axis=1)
    tse = distances[np.arange(datapoints.shape[0]),partition].sum()
    return tse

def compute_CI(centroids, true_centroids):
    """ Calculate centroid index """
    # Calculate distances between centroids
    distances = np.expand_dims(centroids, 1)-true_centroids
    distances = np.linalg.norm(distances, axis=2)
    
    # Check mappings (both way around)
    min_i = np.argmin(distances,axis=0)
    ci1 = 0
    for i in range(len(centroids)):
        if ((min_i == i).sum() == 0):
            ci1 += 1
    min_i = np.argmin(distances,axis=1)
    ci2 = 0
    for i in range(len(centroids)):
        if ((min_i == i).sum() == 0):
            ci2 += 1
        
    return max(ci1, ci2)

class FastRandomSwapCL:
    def __init__(self, init_method="random", num_iters=100, num_kmeans=2, 
                 use_rs = True, use_fast = False):
        self.init_method = init_method
        
        # Number of random swap iterations to do
        self.num_iters = num_iters
        # Number of kmeans updates to do per swap 
        self.num_kmeans = num_kmeans
        # Use random swap
        self.use_rs = use_rs
        # Use fast k-means
        self.use_fast = use_fast
        
        # Create OpenCL context
        self.ctx = cl.create_some_context()
        self.queue = cl.CommandQueue(self.ctx)
    
        self.program = None
        
        self.last_dim = None
        
        self.kernel_source = open(KERNEL_FILE).read()
        
    def _build_kernel(self, data_dim):
        # Build OpenCL kernel (compile the source code to target device)
        # Check if dimensionality has changed since last build. No need to 
        # rebuild if same dimension
        if self.last_dim != data_dim:
            if data_dim == 2:
                self.program = cl.Program(self.ctx, 
                    self.kernel_source.replace("FLOAT_TYPE", "float2")).build()
            elif data_dim == 4:
                self.program = cl.Program(self.ctx, 
                    self.kernel_source.replace("FLOAT_TYPE", "float4")).build()
            elif data_dim == 8:
                self.program = cl.Program(self.ctx, 
                    self.kernel_source.replace("FLOAT_TYPE", "float8")).build()
            else:
                raise ValueError("Dimensionality %d not implemented" % data_dim)
            self.last_dim = data_dim

    def _init_centroids(self, data, num_centroids):
        # Select random datapoints as centroids
        return data[np.random.permutation(np.arange(len(data)))[:num_centroids]]

    def _fit_fkmeans(self, data, num_centroids):
        """ Clustering using fast k-means """
        # --- Following is all kinds of variable inits ---
        # First turn Python values into something OpenCL would understand
        # (i.e. C-style floats, ints and so on
        mf = cl.mem_flags
        my_data = np.array(data, dtype=np.float32)
        centroids = self._init_centroids(data, num_centroids).astype(np.float32)
        partition = np.zeros((len(data),), dtype=np.int32)
        num_dim = np.int32(data.shape[1])
        num_cent = np.int32(num_centroids)
        num_data = np.int32(len(data))
        active_cent = np.ones((num_centroids,), dtype=np.int32)
        min_dist = np.zeros((len(my_data),), dtype=np.float32)
        
        # Compile the kernel
        self._build_kernel(num_dim)
        
        num_kmeans = np.int32(self.num_kmeans)
        
        # Init memory buffers given to workitems
        # These are the parameters sent to OpenCL, and their the memory is 
        # located on device that runs the kernel
        data_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                             hostbuf=my_data)
        cent_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                             hostbuf=centroids)
        part_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                             hostbuf=partition)
        active_cent_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                                    hostbuf=active_cent)
        min_dist_buf  = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                                  hostbuf=min_dist)
        # --- End of initalizing variables/buffers/etc ---
        
        # Run Fast kmeans set amount of iterations
        for i in range(self.num_kmeans):
            # Use fast kmeans partition
            self.program.partition_fast(self.queue, (len(data),), None, 
                        data_buf, cent_buf, num_cent, active_cent_buf,
                        min_dist_buf, part_buf)
            
            # Update centroid locations
            self.program.update_centroids(self.queue, (num_centroids,), 
                        None, 
                        data_buf, cent_buf,
                        active_cent_buf, num_data, part_buf)
        
        # Fetch resulting centroids
        cl.enqueue_copy(self.queue, centroids, cent_buf, is_blocking=True)
        
        # Redo if there were invalid values (e.g. centroid without datapoints)
        if np.any(np.isinf(centroids)):
            return self._fit_rs(data, num_centroids)
        
        return centroids
        
    def _fit_rs(self, data, num_centroids):
        """ Clustering using Random Swap """
        # --- Following is all kinds of variable inits ---
        mf = cl.mem_flags
        my_data = np.array(data, dtype=np.float32)
        centroids = self._init_centroids(data, num_centroids).astype(np.float32)
        partition = np.zeros((len(data),), dtype=np.int32)
        num_dim = np.int32(data.shape[1])
        num_cent = np.int32(num_centroids)
        num_data = np.int32(len(data))
        active_cent = np.ones((num_centroids,), dtype=np.int32)
        min_dist = np.zeros((len(my_data),), dtype=np.float32)
        
        # Compile the kernel
        self._build_kernel(num_dim)
        
        num_kmeans = np.int32(self.num_kmeans)
        old_tse = np.float32(np.inf)
        
        # Init memory buffers given to workitems
        data_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                             hostbuf=my_data)
        cent_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                             hostbuf=centroids)
        part_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR,
                             hostbuf=partition)
        old_cent_buf = cl.Buffer(self.ctx, mf.READ_WRITE, size=centroids.nbytes)
        active_cent_buf = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                                    hostbuf=active_cent)
        old_tse_buf  = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                                 hostbuf=old_tse)
        new_tse_buf  = cl.Buffer(self.ctx, mf.READ_WRITE, size=4)
        min_dist_buf  = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, 
                                  hostbuf=min_dist)
        
        # --- End of initalizing variables/buffers/etc ---
        
        # Repeat random swap for specified many times
        for i in range(self.num_iters):
            # Perform swap op
            # Done in OpenCL to avoid expensive memory transfers
            random_centroid = np.int32(r.randint(0, num_centroids-1))
            random_datapoint = np.int32(r.randint(0, len(data)-1))
            self.program.pre_random_swap(self.queue, (1,), None,
                                         data_buf, cent_buf, old_cent_buf,
                                         active_cent_buf,
                                         num_cent,  random_centroid,
                                         random_datapoint)
            
            # Perform set amoutn of kmeans steps
            for j in range(self.num_kmeans):
                if self.use_fast:
                    # Use fast kmeans partition
                    self.program.partition_fast(self.queue, (len(data),), None, 
                                data_buf, cent_buf, num_cent, active_cent_buf,
                                min_dist_buf, part_buf)
                    
                else:
                    # Use normal partition
                    self.program.partition(self.queue, (len(data),), None, 
                                data_buf, cent_buf, num_cent, min_dist_buf, 
                                part_buf)
                
                # Update centroids
                self.program.update_centroids(self.queue, (num_centroids,), 
                            None, 
                            data_buf, cent_buf,
                            active_cent_buf, num_data, part_buf)
                
            # Calculate new TSE
            self.program.calculate_tse(self.queue, (1,), None,
                                       min_dist_buf, num_data, new_tse_buf)
                                     
            
            # Check if swap was good and update new centroids
            self.program.post_random_swap(self.queue, (1,), None,
                                          cent_buf, old_cent_buf, num_cent,
                                          new_tse_buf, old_tse_buf)
            
            # Make sure all kernels have finished, also does some cleanup
            # which would otherwise result to memory leak
            self.queue.finish()
            
        cl.enqueue_copy(self.queue, centroids, cent_buf, is_blocking=True)
        
        # Redo if there were invalid values
        if np.any(np.isinf(centroids)):
            return self._fit_rs(data, num_centroids)
        
        return centroids
    
    def fit(self, data, num_centroids):
        # Select correct fit method
        if self.use_rs:
            # Use random swap (and possibly fast kmeans)
            return self._fit_rs(data, num_centroids)
        else:
            # Use only fast kmeans
            return self._fit_fkmeans(data, num_centroids)
    
def main(args):
    # Main method
    # Load data
    data = None
    try:
        data = np.loadtxt(args["data"]) 
    except Exception as e:
        print("Could not read data from file "+args["data"])
        exit(1)
    
    # Load true centroids if provided
    true_cent = None
    if args["truecent"] is not None:
        try:
            true_cent = np.loadtxt(args["truecent"]) 
        except Exception as e:
            print("True centroids were provided but could not load file "+
                  args["truecent"])
            exit(1)
    
    num_clusters = args["centroids"]
    if true_cent is not None:
        num_clusters = len(true_cent)
    
    # Create clusterator with given arguments
    clusterator = FastRandomSwapCL(num_iters=args["iterations"],
                                   num_kmeans=args["kmiters"],
                                   use_rs=args["rs"],
                                   use_fast=args["fast"])
    
    # Do clustering
    start_time = time()
    estimated_centroids = clusterator.fit(data, num_clusters)
    time_passed = time() - start_time
    
    # Compute TSE
    TSE = compute_TSE(estimated_centroids, data)
    print("\nTime taken to cluster: %.3f seconds" % time_passed)
    print("TSE: {:>11.6g}".format(TSE))
    
    # Compute CI if true centroids are provided
    if true_cent is not None:
        # Calculate CI
        CI = compute_CI(estimated_centroids, true_cent)
        print("CI: %d" % CI)

    # Output centroid positions if requested
    if args["output"] is not None:
        np.savetxt(args["output"], estimated_centroids)
    

if __name__ == '__main__':
    # Argument parsing
    parser = argparse.ArgumentParser(description="Random Swap and Fast K-means " 
                                                 +"using OpenCL. Expect bugs!"+
                                                 "\nEither -rs or -fast must be"
                                                 +" specified.")
    parser.add_argument('data', type=str, help="Data to be clustered (.txt). "+
                        "Expects data of form NxD, where N = number of points "+
                        "and D = dimensionality")
    parser.add_argument('-t', '--truecent', type=str, 
                        help="True centroid locations (.txt). If provided, "+ 
                        "code will print out CI after clustering.")
    parser.add_argument('-k', '--kmiters', type=int, default=2,
                        help="Number of kmeans iterations per random swap."+
                        "If random swap is not used, this is the maximum "+
                        "number of fast kmeans iterations done")
    parser.add_argument('-d', '--device', type=str, default="",
                        help="OpenCL device to be used. E.g. '0:0' means "+
                        "'first device in first platform'. (See PyOpenCL "+
                        "'create_some_context')")
    parser.add_argument('-i', '--iterations', type=int, default=100,
                        help="Number of random swap iterations")
    parser.add_argument('-c', '--centroids', type=int, default=10,
                        help="Number of centroids. If true centroids are given"+
                        ", this will be set to number of true centroids")
    parser.add_argument('-o', '--output', type=str, default=None,
                        help="File where estimated clusters should be written.")
    parser.add_argument('-fast', action="store_true",
                        help="Use fast-kmeans.")
    parser.add_argument('-rs', action="store_true",
                        help="Use random swap.")
    args = vars(parser.parse_args())
    
    if not args["fast"] and not args["rs"]:
        print("Please select algorithm to use: One or more of -fast or -rs")
        exit(1)
    
    main(args)

