/* NOTE:
   "FLOAT_TYPE" will be replaced with appropiate datatype 
   (float2, float4 or float8) before compiling */

// Partitions data based on clusters
// Global size / number of workers: num_datapoints
__kernel void partition(__global FLOAT_TYPE *data,
                        __global FLOAT_TYPE *centroids,
                                 int        num_cent,
                        __global float      *min_distances,
                        __global int        *partition) {
    uint gid = get_global_id(0);
    
    // Each work-item has one datapoint
    FLOAT_TYPE my_data = data[gid];
    
    // Stores partition
    int my_part = 0;
    // Stores current min distance
    float min_dist = HUGE_VAL;
    // Placeholder for distance calculations
    float dist = 0.0f;
    
    // Loop over centroids
    for (int cent = 0; cent < num_cent; cent++) {
        // Calculate distances between work-item's datapoint
        // and centroids
        dist = length(my_data-centroids[cent]);
        if (dist < min_dist) {
            // Update partition if centroid was closer
            my_part = cent;
            min_dist = dist;
        }
    }
    
    // Store partition and min distance to the global memory
    min_distances[gid] = min_dist;
    partition[gid] = my_part;
}

// Partitions data based on clusters using fast-kmeans partition
// Global size / number of workers: num_datapoints
__kernel void partition_fast(__global FLOAT_TYPE *data,
                             __global FLOAT_TYPE *centroids,
                                      int        num_cent,
                             __global int        *active_centroids,
                             __global float      *min_distances,
                             __global int        *partition) {
    uint gid = get_global_id(0);
    
    // Each work-item has one datapoint
    FLOAT_TYPE my_data = data[gid];
    
    // Stores partition
    int my_part = partition[gid];
    // Stores distance to closest centroid
    float min_dist = min_distances[gid];
    // Placeholder for distance calculations
    float dist = 0.0f;
    
    int is_active = active_centroids[my_part];
        
    // NOTE: This branching is bad for OpenCL!
    if (is_active) {
        // Closest centroid was active, check if we are closer or not
        dist = length(my_data-centroids[my_part]);
        if (dist > min_dist) {
            // Centroid is further away, check against all centroids
            min_dist = dist;
            // Update against all centroids
            for (int cent = 0; cent < num_cent; cent++) {
                dist = length(my_data-centroids[cent]);
                if (dist < min_dist) {
                    my_part = cent;
                    min_dist = dist;
                }
            }
        } else {
            // Else: centroid moved closer, in which case partition is same
            min_dist = dist;
        }
    } else {
        // Closest centroid was not active, check against active centroids
        for (int cent = 0; cent < num_cent; cent++) {
            if (active_centroids[cent] == 0) continue;
            dist = length(my_data-centroids[cent]);
            if (dist < min_dist) {
                my_part = cent;
                min_dist = dist;
            }
        }
    }
    
    // Store partition and min distance to the global memory
    min_distances[gid] = min_dist;
    partition[gid] = my_part;
}

// Update centroids based on the partitions
// Global size / number of workers: num_centroids
__kernel void update_centroids(__global FLOAT_TYPE *data,
                               __global FLOAT_TYPE *centroids,
                               __global int        *active_centroids,
                                        int        len_data,
                               __global int        *partition) {
    uint gid = get_global_id(0);
    
    // Holds number of vectors in this partition
    int part_count = 0;
    // Holds sum of 
    FLOAT_TYPE summed = (FLOAT_TYPE)0.0f;
    
    for (int i = 0; i < len_data; i++) {
        if (partition[i] == gid) {
            summed += data[i];
            part_count += 1;
        }
    }
    if (part_count == 0) {
        // No vectors in this datagroup
        // -> No proper location for centroid
        // -> Set invalid value
        centroids[gid] = (FLOAT_TYPE) HUGE_VAL;
    } else {
        summed = summed/part_count;
        if (any(summed != centroids[gid])) {
            active_centroids[gid] = 1;
        } else {
            active_centroids[gid] = 0;
        }
        centroids[gid] = summed;
    }

}

// Randomly swaps one centroid on one datapoint
// And saves old centroids
// Do it in kernel to avoid reading memory from device to host
// Global size / number of workers: 1
__kernel void pre_random_swap( __global FLOAT_TYPE *data,
                               __global FLOAT_TYPE *centroids,
                               __global FLOAT_TYPE *old_centroids,
                               __global int        *active_centroids,
                                        int        num_cent,
                                        int        random_centroid,
                                        int        random_data) {
    // Update old centroids
    for (int cent = 0; cent < num_cent; cent++) {
        old_centroids[cent] = centroids[cent];
    }
    // Move random centroid on random datapoint
    centroids[random_centroid] = data[random_data];
    // Set moved centroid as an active centroid
    active_centroids[random_centroid] = 1;
}

// Calculate tse and put it in "tse_out"
// Global size / number of workers: 1
__kernel void calculate_tse(__global float      *min_distances,
                                     int        len_data,
                            __global float      *tse_out) {
    float sum_tse = 0.0f;
    for (int i=0; i < len_data; i++) {
        sum_tse += pow(min_distances[i], 2.0f);
    }
    tse_out[0] = sum_tse;
}

// Check if new centroids are better than old ones, 
// "centroids" will be new centroids if new setup is better than previous
// Global size / number of workers: 1
__kernel void post_random_swap(__global FLOAT_TYPE *centroids,
                               __global FLOAT_TYPE *old_centroids,
                                        int        num_cent,
                               __global float      *new_tse,
                               __global float      *old_tse) {
    if (new_tse[0] > old_tse[0]) {
        // Revert new centroids to be old centroids
        for (int cent = 0; cent < num_cent; cent++) {
            centroids[cent] = old_centroids[cent];
        }
    } else {
        // Update old_tse
        old_tse[0] = new_tse[0];
    }
}
